warning off;

var C1            // Household 1 consumption //
    C2            // Household 2 consumption //
    Lam1          // Household 1 stochastic discount factor //
    Lam2          // Household 2 stochastic discount factor //
    infl          // CPI inflation //
    Rd            // %-rate, domestic deposits //
    Rb            // %-rate, domestic bonds//  
    Re            // %-rate paid on reserves in central bank //
    Rtr           // Policy rate //
    w             // Real wage //
    L             // Labor hours or efforts //
    Q             // Domestic bond price //
    Omega varphi  // Variables to simplify notation //
    theta         // Credit shock //
    lam omega     // Lagrange multipliers for commercial bank's constraint //
    re            // Reserves held by commercial bank in central bank //
    d             // Domestic deposits //
    n             // Commercial bank's net worth //
    phr           // Optimal reset price of firms //
    x1 x2         // Variables to simplify notation //
    A             // TFP //
    Y             // Domestic output //
    b             // Supply of domestic bonds //
    bcb           // Demand for domestic bonds by central bank (QE) //
    bfin          // Demand for domestic bonds by commercial bank //
    delta         // Measure of price dispersion //
    C             // Total consumption //
    spread        // Rb minus Rd //
    ;

varexo er eA et eb;

parameters alpha beta1 beta2 chi epsilon gamma infls kappa phi PHI phipi phiy
           rho rhoa rhot sigma sa sr st X Res thetas bcbs;

load param;

model;

//------------------//
// Type 1 Household //
//------------------//

    // (1) -- Stochastic discount factor: //
    
    Lam1 = beta1 * (C1/C1(-1))^(-sigma);

    // (2) -- Euler eq. for deposits: //
    
    1 = Lam1(+1) * infl(+1)^(-1) * Rd;

    // (3) -- Intratemporal FOC: //
    
    C1^(-sigma) * w = phi * L^chi;

//------------------//
// Type 2 Household //
//------------------//

    // (4) -- Stochastic discount factor: //

    Lam2 = beta2 * (C2/C2(-1))^(-sigma);

    // (5) -- Euler eq. for bonds: //

    1 = Lam2(+1) * infl(+1)^(-1) * Rb(+1);

    // (6) -- Return on bonds: //

    Rb = (1 + kappa * Q) / Q(-1);

    // (7) -- Budget constraint: //

    C2 = Q*b;

//------------------------//
// Financial Intermediary //
//------------------------//

    // (8) -- Extra discounting: //

    Omega = 1 - gamma + gamma * theta * varphi;

    // (9) -- Leverage term: //

    varphi = (1 + lam)/theta * Lam1(+1) * Omega(+1) * infl(+1)^(-1) * Rd - (omega * re) / (n * theta);

    // (10) -- FOC wrt bonds: //

    Lam1(+1) * Omega(+1) * infl(+1)^(-1) * (Rb(+1) - Rd) = lam * theta / (1 + lam);

    // (11) -- FOC for reserves: //

    Lam1(+1) * Omega(+1) * infl(+1)^(-1) * (Re - Rd) = -omega / (1 + lam);

//-------//
// Firms //
//-------//

    // (12)-(14) -- Optimal reset price: //
    
    phr = epsilon / (epsilon - 1) * x1/x2;
    x2 = Y + alpha*Lam1(+1)*(infl(+1))^(epsilon-1)*x2(+1);
    x1 = w / A * Y + alpha*Lam1(+1)*(infl(+1))^epsilon*x1(+1);

    // (15) -- Aggregate production function: //

    Y = A * L / delta;

    // (16) -- Price index for domestic goods: //

    1 = (1-alpha)*phr^(1-epsilon) + alpha*infl^(epsilon-1);

    // (17) -- Price dispersion: //

    delta = (1-alpha)*(phr)^(-epsilon) + alpha * (infl)^epsilon * delta(-1);

//--------------//
// Central Bank //
//--------------//

    // (18) -- Balance sheet: //

    Q * bcb = re;

    // (19)-(21) -- Policy rate rule: //

    log(Rtr) = (1-rho)*log(Res) + rho*log(Rtr(-1)) + (1-rho)*(phipi*(log(infl) - log(infls)) + phiy*(log(Y) - log(Y(-1)))) + sr*er;
    Re  = Rtr;
    Rd = Re;

    // (22) -- QE rule: //

    log(bcb) = (1-rho)*log(bcbs) + rho*log(bcb(-1)) + (1-rho)*PHI*(phipi*(log(infl) - log(infls)) + phiy*(log(Y) - log(Y(-1)))) + sr*eb;

//-------------//
// Aggregation //
//-------------//

    // (23) -- Bank's balance sheet: //

    Q*bfin + re = d + n;

    // (24) -- Bank's IC constraint: //

    Q*bfin = varphi * n;

    // (25) -- Bank's aggregate net worth dynamics: //

    n = gamma * infl^(-1) * ((Rb-Rd(-1))*Q(-1)*bfin(-1) + (Re(-1)-Rd(-1))*re(-1) + Rd(-1)*n(-1)) + X;

    // (26) -- Bond market clearing: //

    b = bfin + bcb;

    // (27) -- Aggregate resource constraint: //

    Y = C1 + C2;

//---------------------//
// Stochastic Proceses //
//---------------------//

    // (28) -- TFP: //

    log(A) = rhoa * log(A(-1)) + sa*eA;

    // (29) -- Credit shock: //

    log(theta) = (1-rhot)*log(thetas) + rhot * log(theta(-1)) + st*et;

    // (30) -- Total consumption: //

    C = C1 + C2;

    // (31) -- Interest rate spread: //

    spread = Rb / Rd;

end;

options_.noprint=1;

steady;

shocks;
var er = 1;
var eA = 1;
var et = 1;
var eb = 1;
end;

stoch_simul(order=1,irf=20,ar=0,qz_zero_threshold=1e-8,nodisplay);